@@ -30,6 +30,8 @@ import java.nio.ByteBuffer;
 import java.nio.charset.CharacterCodingException;
 import java.nio.charset.CharsetDecoder;
 import java.nio.charset.CodingErrorAction;
+import java.time.Instant;
+import java.util.Arrays;
 import java.util.Collection;
 import java.util.Iterator;
 import java.util.List;
@@ -108,6 +110,12 @@ public class Variable
             return getValueType().isExtensionType();
         }
 
+        @Override
+        public boolean isTimestampValue()
+        {
+            return false;
+        }
+
         @Override
         public NilValue asNilValue()
         {
@@ -174,6 +182,12 @@ public class Variable
             throw new MessageTypeCastException();
         }
 
+        @Override
+        public TimestampValue asTimestampValue()
+        {
+            throw new MessageTypeCastException();
+        }
+
         @Override
         public boolean equals(Object obj)
         {
@@ -210,7 +224,8 @@ public class Variable
         RAW_STRING(ValueType.STRING),
         LIST(ValueType.ARRAY),
         MAP(ValueType.MAP),
-        EXTENSION(ValueType.EXTENSION);
+        EXTENSION(ValueType.EXTENSION),
+        TIMESTAMP(ValueType.EXTENSION);
 
         private final ValueType valueType;
 
@@ -234,6 +249,7 @@ public class Variable
     private final ArrayValueAccessor arrayAccessor = new ArrayValueAccessor();
     private final MapValueAccessor mapAccessor = new MapValueAccessor();
     private final ExtensionValueAccessor extensionAccessor = new ExtensionValueAccessor();
+    private final TimestampValueAccessor timestampAccessor = new TimestampValueAccessor();
 
     private Type type;
 
@@ -798,6 +814,14 @@ public class Variable
     //
 
     public Variable setArrayValue(List<Value> v)
+    {
+        this.type = Type.LIST;
+        this.accessor = arrayAccessor;
+        this.objectValue = v.toArray();
+        return this;
+    }
+
+    public Variable setArrayValue(Value[] v)
     {
         this.type = Type.LIST;
         this.accessor = arrayAccessor;
@@ -824,29 +848,29 @@ public class Variable
         @Override
         public ImmutableArrayValue immutableValue()
         {
-            return ValueFactory.newArray(list());
+            return ValueFactory.newArray(array());
         }
 
         @Override
         public int size()
         {
-            return list().size();
+            return array().length;
         }
 
         @Override
         public Value get(int index)
         {
-            return list().get(index);
+            return array()[index];
         }
 
         @Override
         public Value getOrNilValue(int index)
         {
-            List<Value> l = list();
-            if (l.size() < index && index >= 0) {
+            Value[] a = array();
+            if (a.length < index && index >= 0) {
                 return ValueFactory.newNil();
             }
-            return l.get(index);
+            return a[index];
         }
 
         @Override
@@ -856,21 +880,21 @@ public class Variable
         }
 
         @Override
-        @SuppressWarnings("unchecked")
         public List<Value> list()
         {
-            return (List<Value>) objectValue;
+            return Arrays.asList(array());
+        }
+
+        public Value[] array()
+        {
+            return (Value[]) objectValue;
         }
 
         @Override
         public void writeTo(MessagePacker pk)
                 throws IOException
         {
-            List<Value> l = list();
-            pk.packArrayHeader(l.size());
-            for (Value e : l) {
-                e.writeTo(pk);
-            }
+            immutableValue().writeTo(pk);
         }
     }
 
@@ -882,7 +906,25 @@ public class Variable
     {
         this.type = Type.MAP;
         this.accessor = mapAccessor;
-        this.objectValue = v;
+        Value[] kvs = new Value[v.size() * 2];
+        Iterator<Map.Entry<Value, Value>> ite = v.entrySet().iterator();
+        int i = 0;
+        while (ite.hasNext()) {
+            Map.Entry<Value, Value> pair = ite.next();
+            kvs[i] = pair.getKey();
+            i++;
+            kvs[i] = pair.getValue();
+            i++;
+        }
+        this.objectValue = kvs;
+        return this;
+    }
+
+    public Variable setMapValue(Value[] kvs)
+    {
+        this.type = Type.MAP;
+        this.accessor = mapAccessor;
+        this.objectValue = kvs;
         return this;
     }
 
@@ -905,66 +947,49 @@ public class Variable
         @Override
         public ImmutableMapValue immutableValue()
         {
-            return ValueFactory.newMap(map());
+            return ValueFactory.newMap(getKeyValueArray());
         }
 
         @Override
         public int size()
         {
-            return map().size();
+            return getKeyValueArray().length / 2;
         }
 
         @Override
         public Set<Value> keySet()
         {
-            return map().keySet();
+            return immutableValue().keySet();
         }
 
         @Override
         public Set<Map.Entry<Value, Value>> entrySet()
         {
-            return map().entrySet();
+            return immutableValue().entrySet();
         }
 
         @Override
         public Collection<Value> values()
         {
-            return map().values();
+            return immutableValue().values();
         }
 
         @Override
         public Value[] getKeyValueArray()
         {
-            Map<Value, Value> v = map();
-            Value[] kvs = new Value[v.size() * 2];
-            Iterator<Map.Entry<Value, Value>> ite = v.entrySet().iterator();
-            int i = 0;
-            while (ite.hasNext()) {
-                Map.Entry<Value, Value> pair = ite.next();
-                kvs[i] = pair.getKey();
-                i++;
-                kvs[i] = pair.getValue();
-                i++;
-            }
-            return kvs;
+            return (Value[]) objectValue;
         }
 
-        @SuppressWarnings("unchecked")
         public Map<Value, Value> map()
         {
-            return (Map<Value, Value>) objectValue;
+            return immutableValue().map();
         }
 
         @Override
         public void writeTo(MessagePacker pk)
                 throws IOException
         {
-            Map<Value, Value> m = map();
-            pk.packArrayHeader(m.size());
-            for (Map.Entry<Value, Value> pair : m.entrySet()) {
-                pair.getKey().writeTo(pk);
-                pair.getValue().writeTo(pk);
-            }
+            immutableValue().writeTo(pk);
         }
     }
 
@@ -1021,6 +1046,86 @@ public class Variable
         }
     }
 
+    public Variable setTimestampValue(Instant timestamp)
+    {
+        this.type = Type.TIMESTAMP;
+        this.accessor = timestampAccessor;
+        this.objectValue = ValueFactory.newTimestamp(timestamp);
+        return this;
+    }
+
+    private class TimestampValueAccessor
+            extends AbstractValueAccessor
+            implements TimestampValue
+    {
+        @Override
+        public boolean isTimestampValue()
+        {
+            return true;
+        }
+
+        @Override
+        public ValueType getValueType()
+        {
+            return ValueType.EXTENSION;
+        }
+
+        @Override
+        public TimestampValue asTimestampValue()
+        {
+            return this;
+        }
+
+        @Override
+        public ImmutableTimestampValue immutableValue()
+        {
+            return (ImmutableTimestampValue) objectValue;
+        }
+
+        @Override
+        public byte getType()
+        {
+            return ((ImmutableTimestampValue) objectValue).getType();
+        }
+
+        @Override
+        public byte[] getData()
+        {
+            return ((ImmutableTimestampValue) objectValue).getData();
+        }
+
+        @Override
+        public void writeTo(MessagePacker pk)
+                throws IOException
+        {
+            ((ImmutableTimestampValue) objectValue).writeTo(pk);
+        }
+
+        @Override
+        public long getEpochSecond()
+        {
+            return ((ImmutableTimestampValue) objectValue).getEpochSecond();
+        }
+
+        @Override
+        public int getNano()
+        {
+            return ((ImmutableTimestampValue) objectValue).getNano();
+        }
+
+        @Override
+        public long toEpochMillis()
+        {
+            return ((ImmutableTimestampValue) objectValue).toEpochMillis();
+        }
+
+        @Override
+        public Instant toInstant()
+        {
+            return ((ImmutableTimestampValue) objectValue).toInstant();
+        }
+    }
+
     ////
     // Value
     //
@@ -1134,6 +1239,12 @@ public class Variable
         return getValueType().isExtensionType();
     }
 
+    @Override
+    public boolean isTimestampValue()
+    {
+        return this.type == Type.TIMESTAMP;
+    }
+
     @Override
     public NilValue asNilValue()
     {
@@ -1232,4 +1343,13 @@ public class Variable
         }
         return (ExtensionValue) accessor;
     }
+
+    @Override
+    public TimestampValue asTimestampValue()
+    {
+        if (!isTimestampValue()) {
+            throw new MessageTypeCastException();
+        }
+        return (TimestampValue) accessor;
+    }
 }
