@@ -25,10 +25,14 @@ import com.fasterxml.jackson.databind.DeserializationFeature;
 import com.fasterxml.jackson.databind.JsonMappingException;
 import com.fasterxml.jackson.databind.KeyDeserializer;
 import com.fasterxml.jackson.databind.ObjectMapper;
+import com.fasterxml.jackson.databind.deser.std.StdDeserializer;
 import com.fasterxml.jackson.databind.module.SimpleModule;
 import org.junit.Test;
 import org.msgpack.core.MessagePack;
 import org.msgpack.core.MessagePacker;
+import org.msgpack.value.ExtensionValue;
+import org.msgpack.value.MapValue;
+import org.msgpack.value.ValueFactory;
 
 import java.io.ByteArrayInputStream;
 import java.io.ByteArrayOutputStream;
@@ -44,10 +48,10 @@ import java.util.Arrays;
 import java.util.HashMap;
 import java.util.List;
 import java.util.Map;
+import java.util.UUID;
 
 import static org.hamcrest.MatcherAssert.assertThat;
 import static org.hamcrest.core.Is.is;
-import static org.hamcrest.core.IsInstanceOf.instanceOf;
 import static org.junit.Assert.assertArrayEquals;
 import static org.junit.Assert.assertEquals;
 import static org.junit.Assert.assertTrue;
@@ -699,34 +703,12 @@ public class MessagePackParserTest
     {
         ByteArrayOutputStream out = new ByteArrayOutputStream();
         MessagePacker packer = MessagePack.newDefaultPacker(out);
-        packer.packArrayHeader(5);
+        packer.packArrayHeader(3);
         // 0: Integer
         packer.packInt(42);
         // 1: String
         packer.packString("foo bar");
-        // 2: ExtensionType(class desr)
-        {
-            TinyPojo t0 = new TinyPojo();
-            t0.t = "t0";
-            TinyPojo t1 = new TinyPojo();
-            t1.t = "t1";
-            NestedListComplexPojo parent = new NestedListComplexPojo();
-            parent.s = "parent";
-            parent.foos = Arrays.asList(t0, t1);
-            byte[] bytes = objectMapper.writeValueAsBytes(parent);
-            packer.packExtensionTypeHeader((byte) 17, bytes.length);
-            packer.addPayload(bytes);
-        }
-        // 3: ExtensionType(type reference deser)
-        {
-            Map<String, Integer> map = new HashMap<String, Integer>();
-            map.put("one", 1);
-            map.put("two", 2);
-            byte[] bytes = objectMapper.writeValueAsBytes(map);
-            packer.packExtensionTypeHeader((byte) 99, bytes.length);
-            packer.addPayload(bytes);
-        }
-        // 4: ExtensionType(custom deser)
+        // 2: ExtensionType
         {
             packer.packExtensionTypeHeader((byte) 31, 4);
             packer.addPayload(new byte[] {(byte) 0xCA, (byte) 0xFE, (byte) 0xBA, (byte) 0xBE});
@@ -734,8 +716,6 @@ public class MessagePackParserTest
         packer.close();
 
         ExtensionTypeCustomDeserializers extTypeCustomDesers = new ExtensionTypeCustomDeserializers();
-        extTypeCustomDesers.addTargetClass((byte) 17, NestedListComplexPojo.class);
-        extTypeCustomDesers.addTargetTypeReference((byte) 99, new TypeReference<Map<String, Integer>>() {});
         extTypeCustomDesers.addCustomDeser((byte) 31, new ExtensionTypeCustomDeserializers.Deser() {
                     @Override
                     public Object deserialize(byte[] data)
@@ -752,27 +732,210 @@ public class MessagePackParserTest
                 new ObjectMapper(new MessagePackFactory().setExtTypeCustomDesers(extTypeCustomDesers));
 
         List<Object> values = objectMapper.readValue(new ByteArrayInputStream(out.toByteArray()), new TypeReference<List<Object>>() {});
-        assertThat(values.size(), is(5));
+        assertThat(values.size(), is(3));
         assertThat((Integer) values.get(0), is(42));
         assertThat((String) values.get(1), is("foo bar"));
+        assertThat((String) values.get(2), is("Java"));
+    }
+
+    static class TripleBytesPojo
+    {
+        public byte first;
+        public byte second;
+        public byte third;
+
+        public TripleBytesPojo(byte first, byte second, byte third)
         {
-            Object v = values.get(2);
-            assertThat(v, is(instanceOf(NestedListComplexPojo.class)));
-            NestedListComplexPojo pojo = (NestedListComplexPojo) v;
-            assertThat(pojo.s, is("parent"));
-            assertThat(pojo.foos.size(), is(2));
-            assertThat(pojo.foos.get(0).t, is("t0"));
-            assertThat(pojo.foos.get(1).t, is("t1"));
+            this.first = first;
+            this.second = second;
+            this.third = third;
         }
+
+        @Override
+        public boolean equals(Object o)
         {
-            Object v = values.get(3);
-            assertThat(v, is(instanceOf(Map.class)));
-            Map<String, Integer> map = (Map<String, Integer>) v;
-            assertThat(map.size(), is(2));
-            assertThat(map.get("one"), is(1));
-            assertThat(map.get("two"), is(2));
+            if (this == o) {
+                return true;
+            }
+            if (!(o instanceof TripleBytesPojo)) {
+                return false;
+            }
+
+            TripleBytesPojo that = (TripleBytesPojo) o;
+
+            if (first != that.first) {
+                return false;
+            }
+            if (second != that.second) {
+                return false;
+            }
+            return third == that.third;
+        }
+
+        @Override
+        public int hashCode()
+        {
+            int result = first;
+            result = 31 * result + (int) second;
+            result = 31 * result + (int) third;
+            return result;
+        }
+
+        @Override
+        public String toString()
+        {
+            // This key format is used when serialized as map key
+            return String.format("%d-%d-%d", first, second, third);
+        }
+
+        static class Deserializer
+                extends StdDeserializer<TripleBytesPojo>
+        {
+            protected Deserializer()
+            {
+                super(TripleBytesPojo.class);
+            }
+
+            @Override
+            public TripleBytesPojo deserialize(JsonParser p, DeserializationContext ctxt)
+                    throws IOException, JsonProcessingException
+            {
+                return TripleBytesPojo.deserialize(p.getBinaryValue());
+            }
+        }
+
+        static class KeyDeserializer
+                extends com.fasterxml.jackson.databind.KeyDeserializer
+        {
+            @Override
+            public Object deserializeKey(String key, DeserializationContext ctxt)
+                    throws IOException
+            {
+                String[] values = key.split("-");
+                return new TripleBytesPojo(
+                        Byte.parseByte(values[0]),
+                        Byte.parseByte(values[1]),
+                        Byte.parseByte(values[2]));
+            }
+        }
+
+        static byte[] serialize(TripleBytesPojo obj)
+        {
+            return new byte[] { obj.first, obj.second, obj.third };
+        }
+
+        static TripleBytesPojo deserialize(byte[] bytes)
+        {
+            return new TripleBytesPojo(bytes[0], bytes[1], bytes[2]);
+        }
+    }
+
+    @Test
+    public void extensionTypeWithPojoInMap()
+            throws IOException
+    {
+        byte extTypeCode = 42;
+
+        ExtensionTypeCustomDeserializers extTypeCustomDesers = new ExtensionTypeCustomDeserializers();
+        extTypeCustomDesers.addCustomDeser(extTypeCode, new ExtensionTypeCustomDeserializers.Deser()
+        {
+            @Override
+            public Object deserialize(byte[] value)
+                    throws IOException
+            {
+                return TripleBytesPojo.deserialize(value);
+            }
+        });
+
+        SimpleModule module = new SimpleModule();
+        module.addDeserializer(TripleBytesPojo.class, new TripleBytesPojo.Deserializer());
+        module.addKeyDeserializer(TripleBytesPojo.class, new TripleBytesPojo.KeyDeserializer());
+        ObjectMapper objectMapper = new ObjectMapper(
+                new MessagePackFactory().setExtTypeCustomDesers(extTypeCustomDesers))
+                .registerModule(module);
+
+        // Prepare serialized data
+        Map<TripleBytesPojo, TripleBytesPojo> originalMap = new HashMap<>();
+        byte[] serializedData;
+        {
+            ValueFactory.MapBuilder mapBuilder = ValueFactory.newMapBuilder();
+            for (int i = 0; i < 4; i++) {
+                TripleBytesPojo keyObj = new TripleBytesPojo((byte) i, (byte) (i + 1), (byte) (i + 2));
+                TripleBytesPojo valueObj = new TripleBytesPojo((byte) (i * 2), (byte) (i * 3), (byte) (i * 4));
+                ExtensionValue k = ValueFactory.newExtension(extTypeCode, TripleBytesPojo.serialize(keyObj));
+                ExtensionValue v = ValueFactory.newExtension(extTypeCode, TripleBytesPojo.serialize(valueObj));
+                mapBuilder.put(k, v);
+                originalMap.put(keyObj, valueObj);
+            }
+            ByteArrayOutputStream output = new ByteArrayOutputStream();
+            MessagePacker packer = MessagePack.newDefaultPacker(output);
+            MapValue mapValue = mapBuilder.build();
+            mapValue.writeTo(packer);
+            packer.close();
+
+            serializedData = output.toByteArray();
+        }
+
+        Map<TripleBytesPojo, TripleBytesPojo> deserializedMap = objectMapper.readValue(serializedData,
+                new TypeReference<Map<TripleBytesPojo, TripleBytesPojo>>() {});
+
+        assertEquals(originalMap.size(), deserializedMap.size());
+        for (Map.Entry<TripleBytesPojo, TripleBytesPojo> entry : originalMap.entrySet()) {
+            assertEquals(entry.getValue(), deserializedMap.get(entry.getKey()));
+        }
+    }
+
+    @Test
+    public void extensionTypeWithUuidInMap()
+            throws IOException
+    {
+        byte extTypeCode = 42;
+
+        ExtensionTypeCustomDeserializers extTypeCustomDesers = new ExtensionTypeCustomDeserializers();
+        extTypeCustomDesers.addCustomDeser(extTypeCode, new ExtensionTypeCustomDeserializers.Deser()
+        {
+            @Override
+            public Object deserialize(byte[] value)
+                    throws IOException
+            {
+                return UUID.fromString(new String(value));
+            }
+        });
+
+        // In this case with UUID, we don't need to add custom deserializers
+        // since jackson-databind already has it.
+        ObjectMapper objectMapper = new ObjectMapper(
+                new MessagePackFactory().setExtTypeCustomDesers(extTypeCustomDesers));
+
+        // Prepare serialized data
+        Map<UUID, UUID> originalMap = new HashMap<>();
+        byte[] serializedData;
+        {
+            ValueFactory.MapBuilder mapBuilder = ValueFactory.newMapBuilder();
+            for (int i = 0; i < 4; i++) {
+                UUID keyObj = UUID.randomUUID();
+                UUID valueObj = UUID.randomUUID();
+                ExtensionValue k = ValueFactory.newExtension(extTypeCode, keyObj.toString().getBytes());
+                ExtensionValue v = ValueFactory.newExtension(extTypeCode, valueObj.toString().getBytes());
+                mapBuilder.put(k, v);
+                originalMap.put(keyObj, valueObj);
+            }
+            ByteArrayOutputStream output = new ByteArrayOutputStream();
+            MessagePacker packer = MessagePack.newDefaultPacker(output);
+            MapValue mapValue = mapBuilder.build();
+            mapValue.writeTo(packer);
+            packer.close();
+
+            serializedData = output.toByteArray();
+        }
+
+        Map<UUID, UUID> deserializedMap = objectMapper.readValue(serializedData,
+                new TypeReference<Map<UUID, UUID>>() {});
+
+        assertEquals(originalMap.size(), deserializedMap.size());
+        for (Map.Entry<UUID, UUID> entry : originalMap.entrySet()) {
+            assertEquals(entry.getValue(), deserializedMap.get(entry.getKey()));
         }
-        assertThat((String) values.get(4), is("Java"));
     }
 
     @Test
